/**
 * Board of Regents of the University of Wisconsin System
 * licenses this file to you under the Apache License,
 * Version 2.0 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a
 * copy of the License at:
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * </p><p>
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 * </p>
 */
package edu.wisc.hrapiclient;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.github.jasminb.jsonapi.ErrorUtils;
import com.github.jasminb.jsonapi.JSONAPIDocument;
import com.github.jasminb.jsonapi.models.errors.Error;
import com.github.jasminb.jsonapi.models.errors.Errors;
import edu.wisc.hrapi.model.AccessToken;
import edu.wisc.hrapi.model.OAuthError;
import edu.wisc.hrapi.model.SupervisoryOrgStats;
import edu.wisc.hrapi.model.SupervisoryOrganization;
import lombok.extern.slf4j.Slf4j;
import okhttp3.Interceptor;
import okhttp3.OkHttpClient;
import okhttp3.Request;
import okhttp3.logging.HttpLoggingInterceptor;
import org.jetbrains.annotations.NotNull;
import retrofit2.Response;

import java.io.IOException;
import java.util.*;

@Slf4j
public class HrAPIClient {

    private final String apikey;
    private final String secret;
    private String bearerToken;
    private long tokenValidUntil=0;

    private boolean enableOkhttpLogging = false;
    private boolean returnNullInsteadOfException = false;

    private OkHttpClient httpClientOauth;
    private OkHttpClient httpClientPersonAPI;

    HrAPIService hrAPIService = null;
    OAuthAPIService oAuthAPIService = null;

    /**
     * Constructor for the PersonAPIClient.
     * <p>
     * The Person API Requires authentication, so the constructor requires the
     * authentication information.
     *
     * @param apiKey The API Key for the registered application
     * @param secret The secret that goes with the API Key
     */
    public HrAPIClient(String apiKey, String secret)
    {
        this.apikey = apiKey;
        this.secret = secret;
    }

    public void shutdown()
    {
        if (httpClientOauth != null)
        {
            httpClientOauth.dispatcher().executorService().shutdown();
            httpClientOauth.connectionPool().evictAll();
            httpClientOauth = null;
        }
        if (httpClientPersonAPI != null)
        {
            httpClientPersonAPI.dispatcher().executorService().shutdown();
            httpClientPersonAPI.connectionPool().evictAll();
            httpClientPersonAPI = null;
        }
    }

    /**
     * To enable debug logging for the okhttp3 layer,
     * call this method immediately after creating the
     * client object. Ths will then add the logging
     * interceptor to the OKHttp client when it is
     * created.
     *
     * @param enable Set to true if you want to enable the http debug logging
     * @since 1.0.0
     */
    public void setEnableHttpLogging(boolean enable)
    {
        if (enable) log.trace("Enableing OkHttpClient intercepters");
        this.enableOkhttpLogging = enable;
        bearerToken = null;
        hrAPIService = null;
        oAuthAPIService = null;
    }

    /**
     * By Default, the client will throw a {@link NotFoundException} if
     * the search result didn't find any results. Some applications
     * would prefer a null or empty list return instead. If you would
     * preffer null returns, call this method with true.
     *
     * @param returnNulls set to true if the client should return null instead of thowing a NotFoundException
     * @since 1.0.0
     */
    public void setReturnNullInsteadOfException(boolean returnNulls)
    {
        this.returnNullInsteadOfException = returnNulls;
    }


    /**
     * Checks the status of the client library if we have
     * a successful authentication to the OAuth service
     * and the token hasn't expired.
     *
     * @return Returns true if the authentication is currently valid
     * @since 1.0.0
     */
    public boolean isAuthenticated()
    {
        return bearerToken != null && tokenValidUntil > System.currentTimeMillis();
    }

    /**
     * This method will call the OAuth authentication endpoint to
     * get the access token used to make person api calls.
     * <p>
     * This can be used by applications that use this library to test
     * if their supplied credentials are valid. The library itself
     * will always check if our access token is still valid and
     * fetch a new one if needed before calling the person api endpoint.
     * </p>
     *
     * @throws AuthenticationException Thrown if there was an issue authenticating
     * @since 1.0.0
     */
    public void authenticate() throws AuthenticationException
    {
        // Authentication is OAuth (kind of) where we post our
        // credentials to the token endpoint /oauth/token,
        // and we get back the AccessToken JSON structure
        // containing the access_token we use to call
        // all the rest of the APIs with in the headers
        // of our call

        // Body of our message will contain....
        Map<String, String> map = new HashMap<>();
        map.put("grant_type","client_credentials");
        map.put("client_id",  apikey);
        map.put("client_secret", secret);

        if (oAuthAPIService == null) {
            log.trace("Creating OAuthAPIService client");
            OkHttpClient.Builder httpClientBuilder  = new OkHttpClient.Builder();
            HttpLoggingInterceptor logging = new HttpLoggingInterceptor().setLevel(HttpLoggingInterceptor.Level.BASIC);

            if (httpClientOauth != null)
            {
                httpClientOauth.dispatcher().executorService().shutdown();
                httpClientOauth.connectionPool().evictAll();
                httpClientOauth = null;
            }

            if (enableOkhttpLogging) {
                if (!httpClientBuilder.interceptors().contains(logging)) {
                    httpClientBuilder.addInterceptor(logging);
                }
            }
            httpClientOauth = httpClientBuilder.build();

            oAuthAPIService = OAuthAPIServiceGenerator.createService(OAuthAPIService.class, httpClientOauth);
        }
        try {
            log.debug("Requesting authentication token");
            Response<AccessToken> tokenRespopnse = oAuthAPIService.getAccessToken(map).execute();

            if (!tokenRespopnse.isSuccessful()) {
                log.debug("Failed in OAuth Authentication token request");
                ObjectMapper mapper = new ObjectMapper();
                OAuthError error = mapper.readValue(Objects.requireNonNull(tokenRespopnse.errorBody()).string(), OAuthError.class);

                hrAPIService = null;
                bearerToken = null;
                tokenValidUntil = 0;
                throw new AuthenticationException(error.getError(), error.getError_description());
            }

            if (tokenRespopnse.body() == null)
                throw new AuthenticationException("error", "Unexpected response from API");

            // Store our info
            bearerToken = tokenRespopnse.body().getAccess_token();
            tokenValidUntil = System.currentTimeMillis() + (tokenRespopnse.body().getExpires_in() * 1000L);

            // Create the personAPI Service
            log.trace("Creating PersonAPIService client");

            OkHttpClient.Builder httpClientBuilder  = new OkHttpClient.Builder();
            HttpLoggingInterceptor logging = new HttpLoggingInterceptor().setLevel(HttpLoggingInterceptor.Level.BASIC);

            if (httpClientPersonAPI != null)
            {
                httpClientPersonAPI.dispatcher().executorService().shutdown();
                httpClientPersonAPI.connectionPool().evictAll();
                httpClientPersonAPI = null;
            }

            httpClientBuilder.interceptors().clear();
            if (enableOkhttpLogging)
                httpClientBuilder.addInterceptor(logging);
            httpClientBuilder.addInterceptor(new Interceptor() {
                @Override
                public @NotNull okhttp3.Response intercept(Interceptor.@NotNull Chain chain) throws IOException {
                    Request original = chain.request();
                    Request.Builder builder = original.newBuilder().header("Authorization", "Bearer " + bearerToken);
                    Request request = builder.build();
                    return chain.proceed(request);
                }
            });
            httpClientPersonAPI = httpClientBuilder.build();

            hrAPIService = HrAPIServiceGenerator.createService(HrAPIService.class, httpClientPersonAPI);
        }
        catch (IOException io)
        {
            throw new AuthenticationException("IO Error", io.getMessage());
        }
    }

    /**
     * The HR API doesn't have any filters on the main endpoint
     * so we are providing a way to walk the way down the /supervisoryOrganizations
     * endpoint.
     * <p>
     * Using this call will return the current info
     * from the endpoint of how many entries exist, how many pages
     * it will take to fetch all the data, and the page size.
     * The current API does not allow the page size to be altered.
     * </p>
     *
     * @return Returns a {@link SupervisoryOrgStats}
     * @throws HrAPIException Thrown when an error from the HR API is returned
     * @throws AuthenticationException Thrown when there was an authentication error
     */
    public SupervisoryOrgStats getSupervisoryOrgStatistics() throws HrAPIException, AuthenticationException
    {
        SupervisoryOrgStats result = new SupervisoryOrgStats();

        if (!isAuthenticated())
            authenticate();

        try {
            // Make a call for page 1 (no options) as we really just want the metadata
            Map<String, String> options = new HashMap<>();
            Response<JSONAPIDocument<List<SupervisoryOrganization>>> response;
            response = hrAPIService.getSupervisoryOrganizations(options).execute();

            if (response.isSuccessful()) {
                if (response.body() != null && response.body().getMeta() != null) {
                    Object tmp = response.body().getMeta().get("paging");
                    if (tmp != null) {
                        Map<String, Integer> paging = (Map<String, Integer>) tmp;
                        result.setPageSize(paging.get("pageSize"));
                        result.setTotalCount(paging.get("totalCount"));
                        result.setPageCount(paging.get("totalPages"));
                        return result;
                    }
                }
                if (returnNullInsteadOfException)
                    return null;
                throw new NotFoundException("No results");

            } else {
                if (response.errorBody() != null) {
                    Errors error = ErrorUtils.parseErrorResponse(new ObjectMapper(), response.errorBody(), Errors.class);
                    throw new HrAPIException(error.getErrors());
                } else {
                    Error e = new Error();
                    e.setDetail("Unknown Error");
                    throw new HrAPIException(e);
                }
            }
        }
        catch (IOException ioException) {
            Error e = new Error();
            e.setDetail(ioException.getMessage());
            e.setCode("IO Error");
            throw new HrAPIException(e);
        }

    }

    /**
     * Fetch the list of supervisory organizations one page at a time
     * <br>
     * The current API does not have any filter for the "get all" call.
     * The only way to get data from this is to fetch it a page at a time.
     *
     * @param pageNumber The page number to fetch from the HR API
     * @return returns a {@link List} of {@link SupervisoryOrganization}
     * @throws NotFoundException thrown when no data is returned
     * @throws HrAPIException Thrown if the HR API returns an error
     * @throws AuthenticationException Thrown if there was an authentication error
     * @since 1.0.0
     */
    public List<SupervisoryOrganization> getSupervisoryOrgPage(long pageNumber) throws NotFoundException, HrAPIException, AuthenticationException
    {
        if (!isAuthenticated())
            authenticate();

        try {

            Map<String, String> options = new HashMap<>();
            options.put("page[number]", ""+pageNumber);
            Response<JSONAPIDocument<List<SupervisoryOrganization>>> response;
            response = hrAPIService.getSupervisoryOrganizations(options).execute();

            if (response.isSuccessful()) {

                // Add our results to the list we will return
                if (response.body() != null && response.body().get() != null) {
                    return Objects.requireNonNull(response.body().get());
                }
                if (returnNullInsteadOfException)
                    return null;
                throw new NotFoundException("No results");
            } else {
                if (response.errorBody() != null) {
                    Errors error = ErrorUtils.parseErrorResponse(new ObjectMapper(), response.errorBody(), Errors.class);
                    throw new HrAPIException(error.getErrors());
                } else {
                    Error e = new Error();
                    e.setDetail("Unknown Error");
                    throw new HrAPIException(e);
                }
            }
        }
        catch (IOException ioException)
        {
            Error e = new Error();
            e.setDetail(ioException.getMessage());
            e.setCode("IO Error");
            throw new HrAPIException(e);
        }
    }

    /**
     * Get the data about a specified Supervisory Org
     *
     * @param supervisoryOrgId The supervisory Org ID to lookup
     * @param  includeSuperiorOrganization should the result include full details about the superior organization
     * @param  includeSubordinateOrganizations should the result include full details about the subordinate organizations
     * @param includeAllSubordinateOrganizations  should the result include full details about the all subordinate orgainzations
     * @return Returns {@link SupervisoryOrganization} found
     * @throws NotFoundException Thrown if no entires found
     * @throws HrAPIException Thrown if the HR API returned an error
     * @throws AuthenticationException Thrown for authentication exceptions
     * @since 1.0.0
     */
    public SupervisoryOrganization getSupervisoryOrg(String supervisoryOrgId, boolean includeSuperiorOrganization, boolean includeSubordinateOrganizations, boolean includeAllSubordinateOrganizations) throws NotFoundException, HrAPIException, AuthenticationException
    {
        // Build our request options
        Map<String, String> options = new HashMap<>();
        String includes = "";
        if (includeSuperiorOrganization)
            includes = "superiorOrganization";
        if (includeSubordinateOrganizations) {
            if (includes.isEmpty())
                includes = "subordinateOrganizations";
            else
                includes += ",subordinateOrganizations";
        }
        if (includeAllSubordinateOrganizations) {
            if (includes.isEmpty())
                includes = "allSubordinateOrganizations";
            else
                includes += ",allSubordinateOrganizations";
        }
        if (!includes.isEmpty())
            options.put("include", includes);

        return getSupervisoryOrgWithFilter(supervisoryOrgId, options);
    }


    private SupervisoryOrganization getSupervisoryOrgWithFilter(String supervisoryOrg, Map<String, String> filter) throws NotFoundException, HrAPIException, AuthenticationException
    {
        if (!isAuthenticated())
            authenticate();

        // Make the call, wrap the IOException that
        // could happen and throw a HrAPIExecption instead
        try {

            Response<JSONAPIDocument<SupervisoryOrganization>> response;
            response = hrAPIService.getSupervisoryOrganizations(supervisoryOrg, filter).execute();

            if (response.isSuccessful()) {

                // Add our results to the list we will return
                if (response.body() != null && response.body().get() != null) {
                    return response.body().get();
                }

            } else {
                if (response.errorBody() != null) {
                    Errors error = ErrorUtils.parseErrorResponse(new ObjectMapper(), response.errorBody(), Errors.class);
                    throw new HrAPIException(error.getErrors());
                } else {
                    Error e = new Error();
                    e.setDetail("Unknown Error");
                    throw new HrAPIException(e);
                }
            }
        }
        catch (IOException ioException)
        {
            Error e = new Error();
            e.setDetail(ioException.getMessage());
            e.setCode("IO Error");
            throw new HrAPIException(e);
        }
        if (returnNullInsteadOfException)
            return null;
        throw new NotFoundException("No results");
    }

}
